import click


def get_repos(labels):
    if not labels:
        click.echo(
            'You need to specify the "labels" object in the config file',
            error=True
        )
        click.exit()
    repos = labels.get('repositories')
    if not isinstance(repos, list):
        click.echo(
            'You need to specify the "repos" list in the config '
            'labels object file',
            error=True
        )
        click.exit()
    return repos



import os
import logging
import configparser

import requests

from .repo import GitRepo
from .components import GithubException


LOGGER = logging.getLogger('github')


class GithubApi:

    def __init__(self, auth=None, http=None):
        self.http = requests.Session()
        try:
            self.auth = auth or get_auth()
        except GithubException as exc:
            LOGGER.warning(str(exc))
            self.auth = None

    @property
    def api_url(self):
        return 'https://api.github.com'

    @property
    def uploads_url(self):
        return 'https://uploads.github.com'

    def __repr__(self):
        return self.api_url
    __str__ = __repr__

    def repo(self, repo_path):
        return GitRepo(self, repo_path)


def get_auth():
    """Return a tuple for authenticating a user

    If not successful raise ``AgileError``.
    """
    auth = get_auth_from_env()
    if auth[0] and auth[1]:
        return auth

    home = os.path.expanduser("~")
    config = os.path.join(home, '.gitconfig')
    if not os.path.isfile(config):
        raise GithubException('No .gitconfig available')

    parser = configparser.ConfigParser()
    parser.read(config)
    if 'user' in parser:
        user = parser['user']
        if 'username' not in user:
            raise GithubException('Specify username in %s user '
                                  'section' % config)
        if 'token' not in user:
            raise GithubException('Specify token in %s user section'
                                  % config)
        return user['username'], user['token']
    else:
        raise GithubException('No user section in %s' % config)


def get_auth_from_env():
    return (
        os.environ.get('GITHUB_USERNAME', ''),
        os.environ.get('GITHUB_TOKEN', '')
    )



import socket

import click

from kong.client import Kong, KongError

from .repo import RepoManager
from .import utils


@click.command()
@click.pass_context
@click.option('--namespace', default='dev', help='target namespace')
@click.option(
    '--yes', is_flag=True,
    help='commit changes to kong', default=False)
def kong(ctx, namespace, yes):
    """Update the kong configuration
    """
    m = KongManager(ctx.obj['agile'], namespace=namespace)
    click.echo(utils.niceJson(m.create_kong(yes)))


class KongManager(RepoManager):

    def create_kong(self, yes):
        data = self.load_data('values.yaml')
        values = data.copy()
        manifest = self.manifest(values, 'kong.yaml')
        if self.namespace == 'local':
            ip = socket.gethostbyname(socket.gethostname())
            for srv in manifest.get('services') or ():
                config = srv.get('config')
                if config:
                    config['host'] = ip
        if yes:
            return self.wait(self.apply_kong(manifest))
        else:
            return manifest

    async def apply_kong(self, manifest):
        async with Kong() as cli:
            try:
                result = await cli.apply_json(manifest)
            except KongError as exc:
                raise utils.CommandError(f'Kong Error: {exc}') from None
            return result



import click

from ..utils import niceJson
from ..repo import RepoManager


@click.command()
@click.pass_context
@click.option(
    '--yes', is_flag=True,
    help='Commit changes to github', default=False)
@click.option(
    '--latest', is_flag=True,
    help='Show latest release in github', default=False)
def release(ctx, yes, latest):
    """Create a new release in github
    """
    m = RepoManager(ctx.obj['agile'])
    api = m.github_repo()
    if latest:
        latest = api.releases.latest()
        if latest:
            click.echo(latest['tag_name'])
    elif m.can_release('stage'):
        branch = m.info['branch']
        version = m.validate_version()
        name = 'v%s' % version
        body = ['Release %s from agiletoolkit' % name]
        data = dict(
            tag_name=name,
            target_commitish=branch,
            name=name,
            body='\n\n'.join(body),
            draft=False,
            prerelease=False
        )
        if yes:
            data = api.releases.create(data=data)
            m.message('Successfully created a new Github release')
        click.echo(niceJson(data))
    else:
        click.echo('skipped')



import os
import re
import asyncio
from typing import Dict
from dataclasses import dataclass
from urllib.parse import urlparse

import yaml

from jinja2 import Template

from .api import GithubApi, GithubException
from .slack import SlackIntegration
from . import utils


NAMESPACES = {'local', 'dev', 'stage', 'sandbox', 'production'}


@dataclass
class Branches:
    dev: str = 'master'
    stage: str = 'deploy'


class RepoManager:

    def __init__(self, config=None, path=None, yes=False, namespace=None):
        self.config = config or {}
        self.path = path or os.getcwd()
        self.deploy_path = os.path.join(self.path, 'deploy')
        self.yes = yes
        self.message_brokers = []
        self.namespace = namespace
        self.info = utils.gitrepo(self.path)
        SlackIntegration.add(self)
        self.init()

    def init(self):
        pass

    def software_version(self):
        """Software version

        Requires the Makefile entry

        version:
            ....
        """
        return utils.version()

    def version(self):
        """Software version of the current repository
        """
        branches = self.branches()
        if self.info['branch'] == branches.stage:
            try:
                return self.software_version()
            except Exception as exc:
                raise utils.CommandError(
                    'Could not obtain repo version, do you have a makefile '
                    'with version entry?\n%s' % exc
                )
        else:
            branch = self.info['branch'].lower()
            branch = re.sub('[^a-z0-9_-]+', '-', branch)
            return f"{branch}-{self.info['head']['id'][:8]}"

    def validate_version(self, prefix='v'):
        """Validate version by checking if it is a valid semantic version
        and its value is higher than latest github tag
        """
        version = self.software_version()
        repo = self.github_repo()
        repo.releases.validate_tag(version, prefix)
        return version

    def skip_build(self):
        """Check if build should be skipped
        """
        skip_msg = self.config.get('skip', '[ci skip]')
        return (
            os.environ.get('CODEBUILD_BUILD_SUCCEEDING') == '0' or
            self.info['current_tag'] or
            skip_msg in self.info['head']['message']
        )

    def can_release(self, target=None):
        if self.skip_build() or self.info["pr"]:
            return False

        branch = self.info['branch']
        branches = self.branches()

        if branch not in (branches.dev, branches.stage):
            return False

        if target and branch != getattr(branches, target):
            return False

        if not target:
            target = 'stage' if branch == branches.stage else 'dev'

        if target == 'stage':
            try:
                self.validate_version()
            except GithubException:
                return False

        return True

    def message(self, msg):
        """Send a message to third party applications
        """
        for broker in self.message_brokers:
            try:
                broker(msg)
            except Exception as exc:
                utils.error(exc)

    def github(self):
        return GithubApi()

    def branches(self):
        return self.get('branches', Branches)

    def target_from_branch(self):
        branches = self.branches()
        return 'dev' if self.info['branch'] == branches.dev else 'stage'

    def github_repo(self) -> str:
        url = self.info['remotes'][0]['url']
        if url.startswith('git@'):
            url = url.split(':')
            assert url[0] == 'git@github.com'
            path = url[1]
        else:
            path = urlparse(url).path[1:]
        bits = path.split('.')
        bits.pop()
        return self.github().repo('.'.join(bits))

    def get(self, name: str, DataClass: type) -> object:
        cfg = self.config.get(name) or {}
        return DataClass(**cfg)

    def load_data(self, *paths: str) -> Dict:
        filename = self.filename(*paths)
        if not os.path.isfile(filename):
            raise utils.CommandError('%s file missing' % filename)
        with open(filename, 'r') as fp:
            data = yaml.load(fp) or {}
        data_namespace = data.pop(self.namespace, None)
        for namespace in NAMESPACES:
            data.pop(namespace, None)
        if data_namespace:
            data.update(data_namespace)
        data['namespace'] = self.namespace
        return data

    def manifest(self, values, *paths, filename=None):
        """Load a manifest file and apply template values
        """
        filename = filename or self.filename(*paths)
        with open(filename, 'r') as fp:
            template = Template(fp.read())
        return yaml.load(template.render(values))

    def filename(self, *paths):
        if not os.path.isdir(self.deploy_path):
            raise utils.CommandError(
                "Path '%s' not available" % self.deploy_path
            )
        return os.path.join(self.deploy_path, *paths)

    def copy_env(self, *args, **kw):
        env = os.environ.copy()
        env.update(*args, **kw)
        return env

    def wait(self, coro):
        return asyncio.get_event_loop().run_until_complete(coro)
